/* eslint-disable max-len */
const postcss = require('postcss')
const valueParser = require('postcss-value-parser')

module.exports = postcss.plugin('postcss-annotations-parser', options => {
  options = options || { preserve: true }

  // is this a CSS variable
  function isCssVarClause (value) {
    return !!value.includes('var(')
  }

  // recursive drill in to CSS variable
  function findFallbackRecursively (value) {
    // get the fallback value from the CSS variable clause
    let regex = /var\((?=\S*['-]*)([a-zA-Z'-]+)\W\s*(.+)\)/
    let matches = value.match(regex)
    if (matches) {
      let fallbackValue = matches[2]

      // if we have another CSS variable clause then keep drilling in
      if (isCssVarClause(fallbackValue)) {
        return findFallbackRecursively(fallbackValue)
      } else {
        return matches[2].trim()
      }
    }

    return value
  }

  // parse token(...) containing a var(...) function to a simple form
  function parseTokenFunction (nodes) {
    // remove non-word bits
    nodes.forEach((n, idx) => {
      if (n.type !== 'word') {
        nodes.splice(idx, 1)
      }
    })

    // remove '+' characters from value
    nodes.forEach((n, idx) => {
      if (n.type === 'word') {
        nodes[idx].value = n.value.replace(/\+/g, '')
      }
    })

    return nodes
  }

  return css => {
    css.walkRules(rule => {
      rule.walkDecls(decl => {
        let hasChanges = false
        let declaration = decl

        // if we'd like to preserve the existing declaration as is then keep it and process a clone instead
        if (options.preserve) {
          declaration = decl.clone()
        }

        let parsedValue = valueParser(declaration.value)

        parsedValue.walk(node => {
          // Since we only want to transform var(...) function values,
          // we can ignore anything else.
          if (node.type !== 'function') return
          if (node.value !== 'var' && node.value !== 'token') return

          // calculate how many params the value has
          let numParams
          if (node.nodes) {
            numParams = node.nodes.filter(n => {
              return ['word', 'function', 'string'].some(v => n.type.includes(v))
            }).length
          }

          // if we have less than 2 we don't have a fallback so do nothing
          if (numParams < 2) return

          // parse token(...) functions that contain a var(...) function
          if (node.value === 'token') {
            node.nodes = parseTokenFunction(node.nodes)
          } else {
            // get the string value for this function
            let nodeValue = valueParser.stringify(node)

            // set the node type to generic word
            node.type = 'word'
            // parse down to the fallback value
            node.value = findFallbackRecursively(nodeValue)
          }

          hasChanges = true
        })

        // set the declaration's value to the newly parsed one
        declaration.value = parsedValue

        // if there were changes and we'd like to preserve the original declaration then insert the new declaration before the existing one
        if (options.preserve && hasChanges) {
          decl.before(declaration)
        }
      })
    })
  }
})
